import httpx
import pytest
from pytest_httpx import HTTPXMock

from kognic.base_clients.cloud_storage.download_handler import DownloadHandler

pytest_plugins = ("pytest_asyncio",)


@pytest.mark.asyncio
class TestAsyncDownloadHandler:
    async def test_download_json_200(self, httpx_mock: HTTPXMock, get_spec):
        httpx_mock.add_response(method="GET", status_code=200, url=get_spec.url, content=get_spec.result_bytes)
        async_download_handler = DownloadHandler()
        json_response = await async_download_handler.get_json(get_spec.url)
        assert json_response == get_spec.result_json
        assert len(httpx_mock.get_requests()) == 1

    async def test_download_json_400(self, httpx_mock: HTTPXMock, get_spec):
        httpx_mock.add_response(method="GET", status_code=400, url=get_spec.url)
        async_download_handler = DownloadHandler()
        with pytest.raises(httpx.HTTPStatusError) as exc_info:
            await async_download_handler.get_json(get_spec.url)
        assert exc_info.value.response.status_code == 400
        assert len(httpx_mock.get_requests()) == 1

    async def test_download_json_500(self, httpx_mock: HTTPXMock, get_spec):
        httpx_mock.add_response(method="GET", status_code=500, url=get_spec.url)
        async_download_handler = DownloadHandler(max_retry_attempts=3, max_retry_wait_time=0.1)
        with pytest.raises(httpx.HTTPStatusError) as exc_info:
            await async_download_handler.get_json(get_spec.url)
        assert exc_info.value.response.status_code == 500
        assert len(httpx_mock.get_requests()) == 4

    async def test_download_json_500_then_200(self, httpx_mock: HTTPXMock, get_spec):
        httpx_mock.add_response(method="GET", status_code=500, url=get_spec.url)
        httpx_mock.add_response(method="GET", status_code=500, url=get_spec.url)
        httpx_mock.add_response(method="GET", status_code=200, url=get_spec.url, content=get_spec.result_bytes)
        async_download_handler = DownloadHandler(max_retry_attempts=3, max_retry_wait_time=0.1)
        json_response = await async_download_handler.get_json(get_spec.url)
        assert json_response == get_spec.result_json
        assert len(httpx_mock.get_requests()) == 3



from typing import Dict, Optional, Union

from pydantic import BaseModel, ConfigDict

try:
    from pydantic.alias_generators import to_camel

    class BaseSerializer(BaseModel):
        model_config = ConfigDict(populate_by_name=True, alias_generator=to_camel, from_attributes=True)

        @classmethod
        def from_json(cls, js: Dict):
            return cls.model_validate(js)

        def to_dict(self, by_alias=True) -> Dict:
            return self.model_dump(exclude_none=True, by_alias=by_alias)

except ImportError:
    # We are on pydantic < 2
    import warnings

    from humps import camelize as to_camel

    warnings.warn(
        "Support for pydantic<2 has been deprecated and will be removed in the next major version. Please update to pydantic>2",
        DeprecationWarning,
    )

    class BaseSerializer(BaseModel):
        @classmethod
        def from_json(cls, js: Dict):
            return cls.parse_obj(js)

        def to_dict(self, by_alias=True) -> Dict:
            return self.dict(exclude_none=True, by_alias=by_alias)

        class Config:
            alias_generator = to_camel
            allow_population_by_field_name = True
            orm_mode = True


CursorIdType = Union[int, str]


class PageMetadata(BaseSerializer):
    next_cursor_id: Optional[CursorIdType]


class PaginatedResponse(BaseSerializer):
    data: Union[dict, list]
    metadata: PageMetadata



"""Client for communicating with the Kognic platform."""
import logging
import urllib.parse
from typing import Optional, Union

from humps import decamelize
from kognic.auth.requests.auth_session import RequestsAuthSession
from requests import HTTPError

from kognic.base_clients import __version__
from kognic.base_clients.models import PaginatedResponse
from kognic.base_clients.retry_support import request_with_retry
from kognic.base_clients.util import filter_none

log = logging.getLogger(__name__)

ENVELOPED_JSON_TAG = "data"


class HttpClient:
    """Http Client dealing with auth and communication with API."""

    def __init__(self, auth, host: str, auth_host: str, client_organization_id: int = None, timeout: int = 60, max_retries: int = 10):
        """
        :param auth: auth credentials, see https://developers.kognic.com/docs/kognic-auth
        :param host: override for api url
        :param auth_host: override for authentication url
        :param client_organization_id: Overrides your users organization id. Only works with a Kognic user.
        :param max_retries: Max number of attempts to retry uploading a file to GCS.
        :param timeout: Max time to wait for response from server.
        """

        self.host = host
        self._auth_req_session = RequestsAuthSession(host=auth_host, auth=auth)
        self.headers = {
            "Accept-Encoding": "gzip",
            "Accept": "application/json",
            "User-Agent": f"kognic-{decamelize(self.__class__.__name__)}/{__version__}",
        }
        self.dryrun_header = {"X-Dryrun": ""}
        self.timeout = timeout
        self.max_retries = max_retries

        if client_organization_id is not None:
            self.headers["X-Organization-Id"] = str(client_organization_id)
            log.warning(
                f"WARNING: You will now act as if you are part of organization: {client_organization_id}. "
                f"This will not work unless you are an Kognic user."
            )

    @property
    def session(self):
        return self._auth_req_session.session

    @staticmethod
    def _http_error_handler(exception: HTTPError) -> None:
        """
        Raises a RuntimeError using error details from an enveloped HTTP 400 response.
        """
        if exception.response is not None and exception.response.status_code == 400:
            try:
                message = exception.response.json()["message"]
            except ValueError:
                message = exception.response.text
            raise RuntimeError(message) from exception
        raise exception from None

    @staticmethod
    def _unwrap_enveloped_json(js: dict) -> Union[dict, list, PaginatedResponse]:
        if isinstance(js, list):
            return js
        elif js is not None and js.get("metadata") is not None:
            return PaginatedResponse.from_json(js)
        elif js is not None and js.get(ENVELOPED_JSON_TAG) is not None:
            return js[ENVELOPED_JSON_TAG]
        return js

    def get(self, endpoint, **kwargs) -> dict:
        r"""Sends a GET request. Returns :class:`dict` object.

        :param endpoint: endpoint to be appended to `client.host`.
        :param \**kwargs: Optional arguments that ``request`` takes.
        :rtype: dict
        """

        kwargs.setdefault("headers", self.headers)
        kwargs.setdefault("timeout", self.timeout)
        url = urllib.parse.urljoin(self.host, endpoint)
        resp = request_with_retry(self.session.get, self.max_retries, self.timeout, HttpClient._http_error_handler, url=url, **kwargs)
        return self._unwrap_enveloped_json(resp.json())

    def post(self, endpoint, data=None, json=None, dryrun=False, discard_response=False, **kwargs) -> Optional[dict]:
        r"""Sends a POST request. Returns :class:`dict` object.

        :param endpoint: endpoint to be appended to `client.host`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) JSON blob to send
        :param dryrun: (optional) Send a dry-run header with the request.
        :param discard_response: (optional) Ignore the response and return None
        :param json: (optional) json to send in the body of the :class:`Request`.
        :param \**kwargs: Optional arguments that ``request`` takes.
        :rtype: dict
        """

        if dryrun:
            headers = {**self.headers, **self.dryrun_header}
        else:
            headers = {**self.headers}

        kwargs.setdefault("headers", headers)
        kwargs.setdefault("timeout", self.timeout)
        resp = request_with_retry(
            self.session.post,
            self.max_retries,
            self.timeout,
            HttpClient._http_error_handler,
            url=f"{self.host}/{endpoint}",
            data=data,
            json=filter_none(json),
            **kwargs,
        )
        if discard_response:
            return None
        else:
            return self._unwrap_enveloped_json(resp.json())

    def put(self, endpoint, data, **kwargs) -> dict:
        r"""Sends a PUT request. Returns :class:`dict` object.

        :param endpoint: endpoint to be appended to `client.host`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \**kwargs: Optional arguments that ``request`` takes.
        :rtype: dict
        """
        kwargs.setdefault("headers", self.headers)
        kwargs.setdefault("timeout", self.timeout)
        resp = request_with_retry(
            self.session.put,
            self.max_retries,
            self.timeout,
            HttpClient._http_error_handler,
            url=f"{self.host}/{endpoint}",
            data=filter_none(data),
            **kwargs,
        )
        return self._unwrap_enveloped_json(resp.json())

    def delete(self, endpoint, dryrun=False, discard_response=False, **kwargs) -> Optional[dict]:
        r"""Sends a DELETE request. Returns :class:`dict` object.

        :param endpoint: endpoint to be appended to `client.host`.
        :param dryrun: (optional) Send a dry-run header with the request.
        :param discard_response: (optional) Ignore the response and return None
        :param \**kwargs: Optional arguments that ``request`` takes.
        :rtype: Optional[dict]
        """

        headers = {**self.headers, **self.dryrun_header} if dryrun else {**self.headers}

        kwargs.setdefault("headers", headers)
        kwargs.setdefault("timeout", self.timeout)
        url = urllib.parse.urljoin(self.host, endpoint)
        resp = request_with_retry(self.session.delete, self.max_retries, self.timeout, HttpClient._http_error_handler, url=url, **kwargs)

        return None if discard_response else self._unwrap_enveloped_json(resp.json())



import json
import logging
from typing import Dict

import httpx

from kognic.base_clients.async_retry_support import request_with_retry

log = logging.getLogger(__name__)


class DownloadHandler:
    def __init__(self, max_retry_attempts: int = 23, max_retry_wait_time: int = 60, timeout: int = 60) -> None:
        """
        :param max_retry_attempts: Max number of attempts to retry requests to GCS.
        :param max_retry_wait_time:  Max with time before retrying requests to GCS.
        :param timeout: Max time to wait for response from server.
        """
        self.max_num_retries = max_retry_attempts
        self.max_retry_wait_time = max_retry_wait_time  # seconds
        self.timeout = timeout  # seconds

    async def get_json(self, url: str) -> Dict:
        return json.loads(await self._download_file(url, self.max_num_retries))

    async def _download_file(self, url: str, number_of_retries: int) -> bytes:
        """
        Download a json file from cloud storage

        :param url: URL of file to download
        :param number_of_retries: Number of download attempts before we stop trying to download
        :return: JSON deserialized to dictionary
        """

        async with httpx.AsyncClient() as httpx_client:
            resp = await request_with_retry(
                httpx_client.get,
                number_of_retries,
                self.max_retry_wait_time,
                url=url,
                timeout=self.timeout,
            )

        return resp.content



import pytest
from pytest_httpx import HTTPXMock

from kognic.base_clients.cloud_storage import UploadSpec
from kognic.base_clients.cloud_storage.async_file_resource_client import AsyncFileResourceClient

pytest_plugins = ("pytest_asyncio",)


@pytest.mark.asyncio
class TestAsyncFileResourceClient:
    async def test_upload_files_local_200(self, httpx_mock: HTTPXMock, put_spec, put_spec_2):
        upload_specs = {
            "random-data.txt": UploadSpec(
                destination=put_spec.url,
                filename=str(put_spec.path),
                content_type="application/octet-stream",
            ),
            "random-data-2.txt": UploadSpec(
                destination=put_spec_2.url,
                filename=str(put_spec_2.path),
                content_type="application/octet-stream",
            ),
        }

        httpx_mock.add_response(url=put_spec.url, method="PUT", status_code=200, match_content=put_spec.content)
        httpx_mock.add_response(url=put_spec_2.url, method="PUT", status_code=200, match_content=put_spec_2.content)
        resource_client = AsyncFileResourceClient(max_retry_attempts=3, max_retry_wait_time=0.1)

        await resource_client.upload_files(upload_specs=upload_specs)
        assert len(httpx_mock.get_requests()) == 2

    async def test_upload_json_500_then_200(self, httpx_mock: HTTPXMock, put_spec):
        httpx_mock.add_response(url=put_spec.url, method="PUT", status_code=500, match_content=put_spec.content)
        httpx_mock.add_response(url=put_spec.url, method="PUT", status_code=500, match_content=put_spec.content)
        httpx_mock.add_response(url=put_spec.url, method="PUT", status_code=200, match_content=put_spec.content)

        resource_client = AsyncFileResourceClient(max_retry_attempts=3, max_retry_wait_time=0.1)
        with open(put_spec.path, "rb") as f:
            await resource_client.upload_json(f, put_spec.url)
        assert len(httpx_mock.get_requests()) == 3

    async def test_download_json_500_then_200(self, httpx_mock: HTTPXMock, get_spec):
        httpx_mock.add_response(url=get_spec.url, method="GET", status_code=500)
        httpx_mock.add_response(url=get_spec.url, method="GET", status_code=500)
        httpx_mock.add_response(url=get_spec.url, method="GET", status_code=200, content=get_spec.result_bytes)

        resource_client = AsyncFileResourceClient(max_retry_attempts=3, max_retry_wait_time=0.1)
        json_response = await resource_client.get_json(get_spec.url)
        assert json_response == get_spec.result_json
        assert len(httpx_mock.get_requests()) == 3
